/* -*- mode: c; c-file-style: "gnu" -*-
 * mod_dpkg.c -- Dpkg log-coloriser module for CCZE
 * Copyright (C) 2007 arno. <arno.@no-log.org>
 *
 * This file is part of ccze.
 *
 * ccze is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ccze is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ccze.h>
#include <stdlib.h>
#include <string.h>

static void ccze_dpkg_setup (void);
static void ccze_dpkg_shutdown (void);
static int ccze_dpkg_handle (const char *str, size_t length, char **rest);


static pcre2_code *reg_dpkg_status, *reg_dpkg_action, *reg_dpkg_conffile;

static char *
ccze_dpkg_status_process(pcre2_match_data *offsets)
{
  char *date, *state, *pkg, *installed_version;
  size_t l;

  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&state, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&pkg, &l);
  pcre2_substring_get_bynumber (offsets, 4,
                                (unsigned char **)&installed_version, &l);

  ccze_print_date(date);
  ccze_space();
  ccze_addstr(CCZE_COLOR_KEYWORD, "status");
  ccze_space();
  ccze_addstr(CCZE_COLOR_PKGSTATUS, state);
  ccze_space();
  ccze_addstr(CCZE_COLOR_PKG, pkg);
  ccze_space();
  ccze_addstr(CCZE_COLOR_DEFAULT, installed_version);
  ccze_newline();

  pcre2_substring_free(date);
  pcre2_substring_free(state);
  pcre2_substring_free(pkg);
  pcre2_substring_free(installed_version);

  return NULL;
}

static char *
ccze_dpkg_action_process(pcre2_match_data *offsets)
{
  char *date, *action, *pkg, *installed_version, *available_version;
  size_t l;

  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&action, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&pkg, &l);
  pcre2_substring_get_bynumber (offsets, 4,
                                (unsigned char **)&installed_version, &l);
  pcre2_substring_get_bynumber (offsets, 5,
                                (unsigned char **)&available_version, &l);

  ccze_print_date(date);
  ccze_space();
  ccze_addstr(CCZE_COLOR_KEYWORD, action);
  ccze_space();
  ccze_addstr(CCZE_COLOR_PKG, pkg);
  ccze_space();
  ccze_addstr(CCZE_COLOR_DEFAULT, installed_version);
  ccze_space();
  ccze_addstr(CCZE_COLOR_DEFAULT, available_version);
  ccze_newline();

  pcre2_substring_free(date);
  pcre2_substring_free(action);
  pcre2_substring_free(pkg);
  pcre2_substring_free(installed_version);
  pcre2_substring_free(available_version);

  return NULL;
}

static char *
ccze_dpkg_conffile_process(pcre2_match_data *offsets)
{
  char *date, *filename, *decision;
  size_t l;
  /* YYYY-MM-DD HH:MM:SS conffile <filename> <decision> */

  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&filename, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&decision, &l);

  ccze_print_date(date);
  ccze_space();
  ccze_addstr(CCZE_COLOR_KEYWORD, "conffile");
  ccze_space();
  ccze_addstr(CCZE_COLOR_FILE, filename);
  ccze_space();
  ccze_addstr(CCZE_COLOR_KEYWORD, decision);
  ccze_newline();

  pcre2_substring_free(date);
  pcre2_substring_free(filename);
  pcre2_substring_free(decision);

  return NULL;
}


static void
ccze_dpkg_setup (void)
{
  int error;
  size_t errptr;

  /* YYYY-MM-DD HH:MM:SS status <state> <pkg> <installed-version> */
  reg_dpkg_status = pcre2_compile(
      "^([-\\d]{10}\\s[:\\d]{8})\\sstatus\\s(\\S+)\\s(\\S+)\\s(\\S+)$",
      PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);

  /* YYYY-MM-DD HH:MM:SS <action> <pkg> <installed-version> <available-version> */
  reg_dpkg_action = pcre2_compile(
      "^([-\\d]{10}\\s[:\\d]{8})\\s(install|upgrade|remove|purge)\\s(\\S+)\\s(\\S+)\\s(\\S+)$",
      PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);

  /* YYYY-MM-DD HH:MM:SS conffile <filename> <decision> */
  reg_dpkg_conffile = pcre2_compile(
      "^([-\\d]{10}\\s[:\\d]{8})\\sconffile\\s(\\S+)\\s(install|keep)$",
      PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);

}

static void
ccze_dpkg_shutdown (void)
{
  pcre2_code_free (reg_dpkg_status);
  pcre2_code_free (reg_dpkg_action);
  pcre2_code_free (reg_dpkg_conffile);
}

static int
ccze_dpkg_handle (const char *str, size_t length, char **rest)
{
  pcre2_match_data *offsets;
  
  offsets = pcre2_match_data_create (99, NULL);
  if (pcre2_match (reg_dpkg_status, str, length, 0, 0, offsets, NULL) >= 0)
  {
      *rest = ccze_dpkg_status_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
  }

  if (pcre2_match (reg_dpkg_action, str, length, 0, 0, offsets, NULL) >= 0)
  {
      *rest = ccze_dpkg_action_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
  }

  if (pcre2_match (reg_dpkg_conffile, str, length, 0, 0, offsets, NULL) >= 0)
  {
      *rest = ccze_dpkg_conffile_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
  }

  pcre2_match_data_free (offsets);

  return 0;
}

CCZE_DEFINE_PLUGIN (dpkg, FULL, "Coloriser for dpkg logs.");
