/* -*- mode: c; c-file-style: "gnu" -*-
 * mod_httpd.c -- httpd-related colorizers for CCZE
 * Copyright (C) 2002, 2003 Gergely Nagy <algernon@bonehunter.rulez.org>
 *
 * This file is part of ccze.
 *
 * ccze is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ccze is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ccze.h>
#include <stdlib.h>
#include <string.h>

static void ccze_httpd_setup (void);
static void ccze_httpd_shutdown (void);
static int ccze_httpd_handle (const char *str, size_t length, char **rest);

static pcre2_code *reg_httpd_access, *reg_httpd_error;

static ccze_color_t
_ccze_httpd_error (const char *level)
{
  if (strstr (level, "debug") || strstr (level, "info") ||
      strstr (level, "notice"))
    return CCZE_COLOR_DEBUG;
  if (strstr (level, "warn"))
    return CCZE_COLOR_WARNING;
  if (strstr (level, "error") || strstr (level, "crit") ||
      strstr (level, "alert") || strstr (level, "emerg"))
    return CCZE_COLOR_ERROR;
  return CCZE_COLOR_UNKNOWN;
}

static char *
ccze_httpd_access_log_process (pcre2_match_data *offsets)
{
  char *host, *vhost, *user, *date, *full_action, *method, *http_code;
  char *gsize, *other;
  size_t l;

  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&vhost, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&host, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&user, &l);
  pcre2_substring_get_bynumber (offsets, 4, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 5, (unsigned char **)&full_action, &l);
  pcre2_substring_get_bynumber (offsets, 6, (unsigned char **)&method, &l);
  pcre2_substring_get_bynumber (offsets, 7, (unsigned char **)&http_code, &l);
  pcre2_substring_get_bynumber (offsets, 8, (unsigned char **)&gsize, &l);
  pcre2_substring_get_bynumber (offsets, 9, (unsigned char **)&other, &l);

  ccze_addstr (CCZE_COLOR_HOST, vhost);
  ccze_space();
  ccze_addstr (CCZE_COLOR_HOST, host);
  if (host[0])
    ccze_space ();
  ccze_addstr (CCZE_COLOR_DEFAULT, "-");
  ccze_space ();

  ccze_addstr (CCZE_COLOR_USER, user);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_DATE, date);
  ccze_space ();
  
  ccze_addstr (ccze_http_action (method), full_action);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_HTTPCODES, http_code);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_GETSIZE, gsize);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_DEFAULT, other);
  ccze_newline ();
  
  pcre2_substring_free (vhost);
  pcre2_substring_free (host);
  pcre2_substring_free (user);
  pcre2_substring_free (date);
  pcre2_substring_free (method);
  pcre2_substring_free (full_action);
  pcre2_substring_free (http_code);
  pcre2_substring_free (gsize);
  pcre2_substring_free (other);

  return NULL;
}

static char *
ccze_httpd_error_log_process (pcre2_match_data *offsets)
{
  char *date, *level, *msg;
  ccze_color_t lcol;
  size_t l;
  
  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&level, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&msg, &l);

  ccze_addstr (CCZE_COLOR_DATE, date);
  ccze_space ();

  lcol = _ccze_httpd_error (level);
  ccze_addstr (lcol, level);
  ccze_space ();

  ccze_addstr (lcol, msg);

  ccze_newline ();

  pcre2_substring_free (date);
  pcre2_substring_free (level);
  pcre2_substring_free (msg);

  return NULL;
}

static void
ccze_httpd_setup (void)
{
  int error;
  size_t errptr;

  reg_httpd_access = pcre2_compile
    ("^(\\S*)\\s(\\S*)?\\s?-\\s(\\S+)\\s(\\[\\d{1,2}\\/\\S*"
     "\\/\\d{4}:\\d{2}:\\d{2}:\\d{2}.{0,6}[^\\]]*\\])\\s"
     "(\"([^ \"]+)\\s*[^\"]*\")\\s(\\d{3})\\s(\\d+|-)\\s*(.*)$",
     PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);

  reg_httpd_error = pcre2_compile
    ("^(\\[\\w{3}\\s\\w{3}\\s{1,2}\\d{1,2}\\s\\d{2}:\\d{2}:\\d{2}\\s"
     "\\d{4}\\])\\s(\\[\\w*\\])\\s(.*)$",
     PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);
}

static void
ccze_httpd_shutdown (void)
{
  pcre2_code_free (reg_httpd_access);
  pcre2_code_free (reg_httpd_error);
}

static int
ccze_httpd_handle (const char *str, size_t length, char **rest)
{
  pcre2_match_data *offsets;

  offsets = pcre2_match_data_create (99, NULL);
  if (pcre2_match (reg_httpd_access, str, length, 0, 0, offsets, NULL) >= 0)
    {
      *rest = ccze_httpd_access_log_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
    }
  if (pcre2_match (reg_httpd_error, str, length, 0, 0, offsets, NULL) >= 0)
    {
      *rest = ccze_httpd_error_log_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
    }

  pcre2_match_data_free (offsets);

  return 0;
}

CCZE_DEFINE_PLUGIN (httpd, FULL,
		    "Coloriser for generic HTTPD access and error logs.");
