/* -*- mode: c; c-file-style: "gnu" -*-
 * mod_postfix.c -- Postfix colorizer for CCZE
 * Copyright (C) 2003 Gergely Nagy <algernon@bonehunter.rulez.org>
 *
 * This file is part of ccze.
 *
 * ccze is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ccze is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ccze.h>
#include <string.h>
#include <stdlib.h>

#include "ccze-compat.h"

static void ccze_postfix_setup (void);
static void ccze_postfix_shutdown (void);
static int ccze_postfix_handle (const char *str, size_t length, char **rest);

static pcre2_code *reg_postfix;

static int
_ccze_postfix_process_one (const char *s, char **rest)
{
  char *field, *value;
  size_t i;
  
  if (!strchr (s, '='))
    return 1;

  field = strndup (s, strchr (s, '=') - s);
  i = strlen (s);
  i -= strlen (field) + 1;
  value = strndup (&s[strlen (field) + 1], i);
  
  ccze_addstr (CCZE_COLOR_FIELD, field);
  ccze_addstr (CCZE_COLOR_DEFAULT, "=");
  ccze_wordcolor_process_one (value, 1);

  return 0;
}

static char *
ccze_postfix_process (pcre2_match_data *offsets)
{
  char *spoolid, *s, *rest, *tmp;
  int r;
  size_t l;
  
  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&spoolid, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&s, &l);
  pcre2_substring_get_bynumber (offsets, 4, (unsigned char **)&rest, &l);

  ccze_addstr (CCZE_COLOR_UNIQN, spoolid);
  ccze_addstr (CCZE_COLOR_DEFAULT, ": ");

  tmp = ccze_strbrk (s, ',');
  
  do
    {
      r = _ccze_postfix_process_one (tmp, &rest);
      if (r)
	ccze_addstr (CCZE_COLOR_DEFAULT, tmp);
      else
	tmp = ccze_strbrk (NULL, ',');
      if (tmp)
	ccze_addstr (CCZE_COLOR_DEFAULT, ",");
    } while (!r && (tmp != NULL));

  pcre2_substring_free (spoolid);
  pcre2_substring_free (s);
  pcre2_substring_free (rest);
  
  return NULL;
}

static void
ccze_postfix_setup (void)
{
  int error;
  size_t errptr;

  reg_postfix = pcre2_compile
    ("^([\\dA-F]+): ((client|to|message-id|uid|resent-message-id|from)(=.*))",
     PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);
}

static void
ccze_postfix_shutdown (void)
{
  pcre2_code_free (reg_postfix);
}

static int
ccze_postfix_handle (const char *str, size_t length, char **rest)
{
  pcre2_match_data *offsets;
  
  offsets = pcre2_match_data_create (99, NULL);
  if (pcre2_match (reg_postfix, str, length, 0, 0, offsets, NULL) >= 0)
    {
      if (rest)
	*rest = ccze_postfix_process (offsets);
      else
	ccze_postfix_process (offsets);
      
      pcre2_match_data_free (offsets);
      return 1;
    }

  pcre2_match_data_free (offsets);
    
  return 0;
}

CCZE_DEFINE_PLUGIN (postfix, PARTIAL, "Coloriser for postfix(1) sub-logs.");
