/* -*- mode: c; c-file-style: "gnu" -*-
 * mod_squid.c -- Squid-related colorizers for CCZE
 * Copyright (C) 2002, 2003 Gergely Nagy <algernon@bonehunter.rulez.org>
 *
 * This file is part of ccze.
 *
 * ccze is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ccze is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <ccze.h>
#include <string.h>
#include <stdlib.h>

static void ccze_squid_setup (void);
static void ccze_squid_shutdown (void);
static int ccze_squid_handle (const char *str, size_t length, char **rest);

static pcre2_code *reg_squid_access, *reg_squid_store, *reg_squid_cache;

static ccze_color_t
_ccze_proxy_action (const char *action)
{
  if (strstr (action, "ERR") == action)
    return CCZE_COLOR_ERROR;
  if (strstr (action, "MISS"))
    return CCZE_COLOR_PROXY_MISS;
  if (strstr (action, "HIT"))
    return CCZE_COLOR_PROXY_HIT;
  if (strstr (action, "DENIED"))
    return CCZE_COLOR_PROXY_DENIED;
  if (strstr (action, "REFRESH"))
    return CCZE_COLOR_PROXY_REFRESH;
  if (strstr (action, "SWAPFAIL"))
    return CCZE_COLOR_PROXY_SWAPFAIL;
  if (strstr (action, "NONE"))
    return CCZE_COLOR_DEBUG;

  return CCZE_COLOR_UNKNOWN;
}

static ccze_color_t
_ccze_proxy_hierarch (const char *hierar)
{
  if (strstr (hierar, "NO") == hierar)
    return CCZE_COLOR_WARNING;
  if (strstr (hierar, "DIRECT"))
    return CCZE_COLOR_PROXY_DIRECT;
  if (strstr (hierar, "PARENT"))
    return CCZE_COLOR_PROXY_PARENT;
  if (strstr (hierar, "MISS"))
    return CCZE_COLOR_PROXY_MISS;

  return CCZE_COLOR_UNKNOWN;
}

static ccze_color_t
_ccze_proxy_tag (const char *tag)
{
  if (strstr (tag, "CREATE"))
    return CCZE_COLOR_PROXY_CREATE;
  if (strstr (tag, "SWAPIN"))
    return CCZE_COLOR_PROXY_SWAPIN;
  if (strstr (tag, "SWAPOUT"))
    return CCZE_COLOR_PROXY_SWAPOUT;
  if (strstr (tag, "RELEASE"))
    return CCZE_COLOR_PROXY_RELEASE;

  return CCZE_COLOR_UNKNOWN;
}

static char *
ccze_squid_access_log_process (pcre2_match_data *offsets)
{
  char *date, *espace, *elaps, *host, *action, *httpc, *gsize;
  char *method, *uri, *ident, *hierar, *fhost, *ctype;
  size_t l;

  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&espace, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&elaps, &l);
  pcre2_substring_get_bynumber (offsets, 4, (unsigned char **)&host, &l);
  pcre2_substring_get_bynumber (offsets, 5, (unsigned char **)&action, &l);
  pcre2_substring_get_bynumber (offsets, 6, (unsigned char **)&httpc, &l);
  pcre2_substring_get_bynumber (offsets, 7, (unsigned char **)&gsize, &l);
  pcre2_substring_get_bynumber (offsets, 8, (unsigned char **)&method, &l);
  pcre2_substring_get_bynumber (offsets, 9, (unsigned char **)&uri, &l);
  pcre2_substring_get_bynumber (offsets, 10, (unsigned char **)&ident, &l);
  pcre2_substring_get_bynumber (offsets, 11, (unsigned char **)&hierar, &l);
  pcre2_substring_get_bynumber (offsets, 12, (unsigned char **)&fhost, &l);
  pcre2_substring_get_bynumber (offsets, 13, (unsigned char **)&ctype, &l);

  ccze_print_date (date);
  ccze_addstr (CCZE_COLOR_DEFAULT, espace);
  ccze_addstr (CCZE_COLOR_GETTIME, elaps);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_HOST, host);
  ccze_space ();

  ccze_addstr (_ccze_proxy_action (action), action);
  ccze_addstr (CCZE_COLOR_DEFAULT, "/");
  ccze_addstr (CCZE_COLOR_HTTPCODES, httpc);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_GETSIZE, gsize);
  ccze_space ();

  ccze_addstr (ccze_http_action (method), method);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_URI, uri);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_IDENT, ident);
  ccze_space ();

  ccze_addstr (_ccze_proxy_hierarch (hierar), hierar);
  ccze_addstr (CCZE_COLOR_DEFAULT, "/");
  ccze_addstr (CCZE_COLOR_HOST, fhost);
  ccze_space ();

  ccze_addstr (CCZE_COLOR_CTYPE, ctype);

  ccze_newline ();

  pcre2_substring_free (date);
  pcre2_substring_free (espace);
  pcre2_substring_free (elaps);
  pcre2_substring_free (host);
  pcre2_substring_free (action);
  pcre2_substring_free (httpc);
  pcre2_substring_free (gsize);
  pcre2_substring_free (method);
  pcre2_substring_free (uri);
  pcre2_substring_free (ident);
  pcre2_substring_free (hierar);
  pcre2_substring_free (fhost);
  pcre2_substring_free (ctype);
  
  return NULL;
}

static char *
ccze_squid_cache_log_process (pcre2_match_data *offsets)
{
  char *date, *other, *ret;
  size_t l;

  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&other, &l);

  ccze_addstr (CCZE_COLOR_DATE, date);
  ccze_space();

  ret = strdup (other);
  pcre2_substring_free (date);
  pcre2_substring_free (other);
  return ret;
}

static char *
ccze_squid_store_log_process (pcre2_match_data *offsets)
{
  char *date, *tag, *swapnum, *swapname, *swapsum, *space1, *hcode;
  char *hdate, *lmdate, *expire, *ctype, *size, *read, *method;
  char *uri, *space2, *space3, *space4;
  size_t l;
    
  pcre2_substring_get_bynumber (offsets, 1, (unsigned char **)&date, &l);
  pcre2_substring_get_bynumber (offsets, 2, (unsigned char **)&tag, &l);
  pcre2_substring_get_bynumber (offsets, 3, (unsigned char **)&swapnum, &l);
  pcre2_substring_get_bynumber (offsets, 4, (unsigned char **)&swapname, &l);
  pcre2_substring_get_bynumber (offsets, 5, (unsigned char **)&swapsum, &l);
  pcre2_substring_get_bynumber (offsets, 6, (unsigned char **)&space1, &l);
  pcre2_substring_get_bynumber (offsets, 7, (unsigned char **)&hcode, &l);
  pcre2_substring_get_bynumber (offsets, 8, (unsigned char **)&space2, &l);
  pcre2_substring_get_bynumber (offsets, 9, (unsigned char **)&hdate, &l);
  pcre2_substring_get_bynumber (offsets, 10, (unsigned char **)&space3, &l);
  pcre2_substring_get_bynumber (offsets, 11, (unsigned char **)&lmdate, &l);
  pcre2_substring_get_bynumber (offsets, 12, (unsigned char **)&space4, &l);
  pcre2_substring_get_bynumber (offsets, 13, (unsigned char **)&expire, &l);
  pcre2_substring_get_bynumber (offsets, 14, (unsigned char **)&ctype, &l);
  pcre2_substring_get_bynumber (offsets, 15, (unsigned char **)&size, &l);
  pcre2_substring_get_bynumber (offsets, 16, (unsigned char **)&read, &l);
  pcre2_substring_get_bynumber (offsets, 17, (unsigned char **)&method, &l);
  pcre2_substring_get_bynumber (offsets, 18, (unsigned char **)&uri, &l);

  ccze_print_date (date);
  ccze_space();
  ccze_addstr (_ccze_proxy_tag (tag), tag);
  ccze_space();
  ccze_addstr (CCZE_COLOR_SWAPNUM, swapnum);
  ccze_space();
  ccze_addstr (CCZE_COLOR_SWAPNUM, swapname);
  ccze_space();
  ccze_addstr (CCZE_COLOR_SWAPNUM, swapsum);
  ccze_addstr (CCZE_COLOR_DEFAULT, space1);
  ccze_addstr (CCZE_COLOR_HTTPCODES, hcode);
  ccze_addstr (CCZE_COLOR_DEFAULT, space2);
  ccze_print_date (hdate);
  ccze_addstr (CCZE_COLOR_DEFAULT, space3);
  ccze_print_date (lmdate);
  ccze_addstr (CCZE_COLOR_DEFAULT, space4);
  ccze_print_date (expire);
  ccze_space();
  ccze_addstr (CCZE_COLOR_CTYPE, ctype);
  ccze_space();
  ccze_addstr (CCZE_COLOR_GETSIZE, size);
  ccze_addstr (CCZE_COLOR_DEFAULT, "/");
  ccze_addstr (CCZE_COLOR_GETSIZE, read);
  ccze_space();
  ccze_addstr (ccze_http_action (method), method);
  ccze_space();
  ccze_addstr (CCZE_COLOR_URI, uri);

  ccze_newline ();

  pcre2_substring_free (date);
  pcre2_substring_free (tag);
  pcre2_substring_free (swapnum);
  pcre2_substring_free (swapname);
  pcre2_substring_free (swapsum);
  pcre2_substring_free (space1);
  pcre2_substring_free (hcode);
  pcre2_substring_free (hdate);
  pcre2_substring_free (lmdate);
  pcre2_substring_free (expire);
  pcre2_substring_free (ctype);
  pcre2_substring_free (size);
  pcre2_substring_free (read);
  pcre2_substring_free (method);
  pcre2_substring_free (uri);
  pcre2_substring_free (space2);
  pcre2_substring_free (space3);
  pcre2_substring_free (space4);
  
  return NULL;
}

static void
ccze_squid_setup (void)
{
  int error;
  size_t errptr;

  reg_squid_access = pcre2_compile
    ("^(\\d{9,10}\\.\\d{3})(\\s+)(\\d+)\\s(\\S+)\\s(\\w+)\\/(\\d{3})"
     "\\s(\\d+)\\s(\\w+)\\s(\\S+)\\s(\\S+)\\s(\\w+)\\/([\\d\\.]+|-)\\s(.*)",
     PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);

  reg_squid_cache = pcre2_compile
    ("^(\\d{4}\\/\\d{2}\\/\\d{2}\\s(\\d{2}:){2}\\d{2}\\|)\\s(.*)$",
     PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);

  reg_squid_store = pcre2_compile
    ("^([\\d\\.]+)\\s(\\w+)\\s(\\-?[\\dA-F]+)\\s+(\\S+)\\s([\\dA-F]+)"
     "(\\s+)(\\d{3}|\\?)(\\s+)(\\-?[\\d\\?]+)(\\s+)(\\-?[\\d\\?]+)(\\s+)"
     "(\\-?[\\d\\?]+)\\s(\\S+)\\s(\\-?[\\d|\\?]+)\\/(\\-?[\\d|\\?]+)\\s"
     "(\\S+)\\s(.*)", PCRE2_ZERO_TERMINATED, 0, &error, &errptr, NULL);
}

static void
ccze_squid_shutdown (void)
{
  pcre2_code_free (reg_squid_access);
  pcre2_code_free (reg_squid_cache);
  pcre2_code_free (reg_squid_store);
}

static int
ccze_squid_handle (const char *str, size_t length, char **rest)
{
  pcre2_match_data *offsets;

  offsets = pcre2_match_data_create (99, NULL);
  if (pcre2_match (reg_squid_access, str, length, 0, 0, offsets, NULL) >= 0)
    {
      *rest = ccze_squid_access_log_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
    }
  
  if (pcre2_match (reg_squid_store, str, length, 0, 0, offsets, NULL) >= 0)
    {
      *rest = ccze_squid_store_log_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
    }
  
  if (pcre2_match (reg_squid_cache, str, length, 0, 0, offsets, NULL) >= 0)
    {
      *rest = ccze_squid_cache_log_process (offsets);
      pcre2_match_data_free (offsets);
      return 1;
    }

  pcre2_match_data_free (offsets);

  return 0;
}

CCZE_DEFINE_PLUGIN (squid, FULL,
		    "Coloriser for squid access, store and cache logs.");
